import { next, range } from "./utils.ts";

export function it<T>(iter: Iterable<T>) {
	return new Iter(iter);
}

export function* map<I, O>(iter: Iterable<I>, mapping: (item: I) => O): Generator<O, void, void> {
	for (const item of iter) yield mapping(item);
}

export function* filter<I>(iter: Iterable<I>, predicate: (item: I) => boolean): Generator<I, void, void> {
	for (const item of iter) if (predicate(item)) yield item;
}

export function filter_map<I, O>(
	iter: Iterable<I>,
	mapping: (item: I) => O,
): Generator<Exclude<O, null | undefined>, void, void> {
	const filtered = filter(map(iter, mapping), (item) => item !== null && item !== undefined);
	return filtered as Generator<Exclude<O, null | undefined>, void, void>;
}

export function all<I>(iter: Iterable<I>, predicate: (item: I) => boolean) {
	for (const item of iter) if (!predicate(item)) return false;
	return true;
}

export function* enumerate<T>(iterator: Iterable<T>): Generator<[number, T], void, void> {
	let index = 0;
	for (const item of iterator) yield [index++, item];
}

export function* zip<A, B>(a: Iterable<A>, b: Iterable<B>): Generator<[A, B], void, void> {
	const iter_a = a[Symbol.iterator](), iter_b = b[Symbol.iterator]();
	while (true) {
		const next_a = next(iter_a), next_b = next(iter_b);
		if (next_a === null || next_b === null) return;
		yield [next_a, next_b];
	}
}

export class Iter<T> {
	iter;
	[Symbol.iterator] = () => this.iter[Symbol.iterator]();

	constructor(iter: Iterable<T>) {
		this.iter = iter;
	}

	map<O>(mapping: (item: T) => O) {
		return new Iter(map(this, mapping));
	}

	filter(predicate: (item: T) => boolean) {
		return new Iter(filter(this, predicate));
	}

	filter_map<O>(mapping: (item: T) => O) {
		return new Iter(filter_map(this, mapping));
	}

	enumerate() {
		return new Iter(enumerate(this));
	}

	zip<U>(other: Iterable<U>) {
		return new Iter(zip(this, other));
	}

	array() {
		// deno-lint-ignore no-explicit-any
		const result: any = [...this];
		result.it = () => it(result);
		return result as T[] & { it: () => Iter<T> };
	}

	collect<O>(class_: new (it: Iterable<T>) => O): O {
		return new class_(this);
	}

	all(predicate: (item: T) => boolean) {
		return all(this, predicate);
	}
}

Deno.test("iter", async () => {
	const items = it(range(0, 10))
		.map((e) => e * 3)
		.enumerate()
		.filter(([_, e]) => (e % 2) == 1)
		.zip(range(0, 10))
		.map(JSON.stringify)
		.collect(Set<string>);

	const { assertEquals } = await import("https://deno.land/std@0.224.0/assert/assert_equals.ts");
	assertEquals(items, new Set(["[[1,3],0]", "[[3,9],1]", "[[5,15],2]", "[[7,21],3]", "[[9,27],4]"]));
});
